/*
 * Copyright 2010 Christian Buchner
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <GL/glew.h>
#if defined(__APPLE__) || defined(MACOSX)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <cuda_runtime.h>
#include <cutil_inline.h>
#include <cutil_gl_inline.h>
#include <cuda_gl_interop.h>

#include "Buddhabrot_kernels.h"

//
// CUDA Buddhabrot
//
// Use the '-' and '=' keys to change the scale factor.
//
// Other keys:

void cleanup(void);
void initializeData(int width, int height) ;

static char *sSDKsample = "CUDA Buddhabrot"; 

static int wWidth   = 640;  // Window width
static int wHeight  = 480;  // Window height
static int imWidth  = 0;    // Image width
static int imHeight = 0;    // Image height
bool clear = true;          // clear the image flag

int maxR=380;
int maxG=200;
int maxB=1000;
int colormode=1;

bool lmb = false;
int mod = 0;
bool selection = false;
int xctr, yctr;
int xmin, xmax;
int ymin, ymax;

int fpsCount = 0;          // FPS count for averaging
int fpsLimit = 20;         // FPS limit for sampling
unsigned int timer;
unsigned int g_Bpp;

bool g_bFBODisplay = false;

// Display Data
static GLuint pbo_buffer = 0;  // Front and back CA buffers
static GLuint texid = 0;       // Texture for display
unsigned int * pixels = NULL;  // Image pixel data on the host
int   imageOffset = 0;         // Image brightness offset
float imageScale = 5.f;        // Image exposure
bool  imageFilter = false;     // Image denoise filter
float isigma = 10.0f;          // spatial locality factor for mutations:
                               // the visible range zrmin/max and zimin/max gets divided
                               // by isigma and is used as standard deviation for mutation

bool bQuit = false;

#define OFFSET(i) ((char *)NULL + (i))
#define MAX(a,b) ((a > b) ? a : b)

void computeFPS()
{
    fpsCount++;
    if (fpsCount == fpsLimit) {
        char fps[256];
        float ifps = 1.f / (cutGetAverageTimerValue(timer) / 1000.f);
        sprintf(fps, "CUDA Buddhabrot: %3.1f fps", ifps);

        glutSetWindowTitle(fps);
        fpsCount = 0; 

        cutilCheckError(cutResetTimer(timer));
    }
}


// This is the normal display path
void display(void) 
{  
    cutilCheckError(cutStartTimer(timer));

    unsigned int maxiter = MAX(maxR,MAX(maxG,maxB));

    // Buddhabrot rendering
    Pixel *displaydata = NULL;
    cutilSafeCall(cudaGLMapBufferObject((void**)&displaydata, pbo_buffer));
    Buddhabrot(displaydata, imWidth, imHeight, clear, g_Bpp, imageOffset, imageScale, imageFilter, isigma, maxiter, maxR, maxG, maxB, colormode );
    clear = false;
    cutilSafeCall(cudaGLUnmapBufferObject(pbo_buffer));

    glClear(GL_COLOR_BUFFER_BIT);

    glBindTexture(GL_TEXTURE_2D, texid);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo_buffer);
//    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, imWidth, imHeight, 
//                   GL_LUMINANCE, GL_UNSIGNED_INT, OFFSET(0));
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, imWidth, imHeight, 
                   GL_BGRA, GL_UNSIGNED_BYTE, OFFSET(0));
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);

    glDisable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    glBegin(GL_QUADS);
    glTexCoord2f(0, 1); glVertex2f(0, 0);
    glTexCoord2f(0, 0); glVertex2f(0, 1);
    glTexCoord2f(1, 0); glVertex2f(1, 1);
    glTexCoord2f(1, 1); glVertex2f(1, 0);
    glEnd();
    glBindTexture(GL_TEXTURE_2D, 0);

    if (selection)
    {
        glBegin(GL_LINE_STRIP);
        glVertex2f((float)xmin/wWidth, (float)(wHeight-1-ymin)/wHeight);
        glVertex2f((float)xmin/wWidth, (float)(wHeight-1-ymax)/wHeight);
        glVertex2f((float)xmax/wWidth, (float)(wHeight-1-ymax)/wHeight);
        glVertex2f((float)xmax/wWidth, (float)(wHeight-1-ymin)/wHeight);
        glVertex2f((float)xmin/wWidth, (float)(wHeight-1-ymin)/wHeight);
        glEnd();
    }

    glutSwapBuffers();

    cutilCheckError(cutStopTimer(timer));

    computeFPS();

    glutPostRedisplay();
}

void idle(void) {
    glutPostRedisplay();
}

void mouse(int button, int state, int x, int y)
{
    if (state == GLUT_DOWN && button ==  GLUT_LEFT_BUTTON)
    {
        lmb = true;
        mod = glutGetModifiers();
        xctr = x;
        yctr = y;

        selection = true;
        xmin = x;
        ymin = y;
        xmax = x;
        ymax = y;
    }

    if (state == GLUT_UP && button ==  GLUT_LEFT_BUTTON)
    {
        lmb = false;

        selection = true;
        if (mod & GLUT_ACTIVE_CTRL)
        {
            xmax = x;
            xmin = xctr - (xmax-xctr);

            if (mod & GLUT_ACTIVE_SHIFT)
            {
                ymax = y;
                ymin = yctr - (ymax-yctr);
            }
            else
            {
                ymax = yctr + (xmax-xctr)*wHeight/wWidth;
                ymin = yctr - (xmax-xctr)*wHeight/wWidth;
            }
        }
        else
        {
            xmax = x;
            if (mod & GLUT_ACTIVE_SHIFT)
                ymax = y;
            else
                ymax = ymin+(xmax-xmin)*wHeight/wWidth;
        }

        if (xmax < xmin) { int tmp = xmin; xmin = xmax; xmax = tmp; }
        if (ymax < ymin) { int tmp = ymin; ymin = ymax; ymax = tmp; }
    }

    if (state == GLUT_DOWN && button == GLUT_RIGHT_BUTTON)
    {
        if (x >= xmin && x <= xmax && y >= ymin && y <= ymax)
        {
            Zoom(wWidth, wHeight, xmin, xmax, ymin, ymax);
            clear = true;
        }
        selection = false;
    }
}

void motion(int x, int y)
{
    if (lmb)
    {
        if (mod & GLUT_ACTIVE_CTRL)
        {
            xmax = x;
            xmin = xctr - (xmax-xctr);

            if (mod & GLUT_ACTIVE_SHIFT)
            {
                ymax = y;
                ymin = yctr - (ymax-yctr);
            }
            else
            {
                ymax = yctr + (xmax-xctr)*wHeight/wWidth;
                ymin = yctr - (xmax-xctr)*wHeight/wWidth;
            }
        }
        else
        {
            xmax = x;
            if (mod & GLUT_ACTIVE_SHIFT)
                ymax = y;
            else
                ymax = ymin+(xmax-xmin)*wHeight/wWidth;
        }
    }
}

void keyboard( unsigned char key, int /*x*/, int /*y*/) 
{
    switch( key) {
        case 27:
            exit (0);
        break;
        case '-':
            imageScale *= 1.0f/1.1f;
            printf("\nbrightness = %4.2f\n", imageScale);
        break;
        case '+':
            imageScale *= 1.1f;
            printf("\nbrightness = %4.2f\n", imageScale);
        break;
        case '>':
            imageOffset++;
            printf("\noffset = %d\n", imageOffset);
        break;
        case '<':
            imageOffset--;
            printf("\noffset = %d\n", imageOffset);
        break;
        case 'f':
            imageFilter = !imageFilter;
            printf("\nfilter = %d\n", imageFilter);
        break;
        case 'c':
            clear = true;
        break;
        case 'z':
            if (selection)
            {
                Zoom(wWidth, wHeight, xmin, xmax, ymin, ymax);
                selection = false;
                clear = true;
            }
        break;
        case ' ':
                ResetZoom(wWidth, wHeight);
                selection = false;
                clear = true;
        break;
    }
    glutPostRedisplay();
}

void reshape(int x, int y) {
    glViewport(0, 0, x, y);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, 1, 0, 1, 0, 1); 
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glutPostRedisplay();

    wWidth = x; wHeight = y;
    initializeData(wWidth, wHeight);
}

void cleanup(void) {
    cutilSafeCall(cudaGLUnregisterBufferObject(pbo_buffer));

    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
    glDeleteBuffers(1, &pbo_buffer);
    glDeleteTextures(1, &texid);
    deleteTexture();

    cutilCheckError(cutDeleteTimer(timer));  
}

void initializeData(int width, int height) {
    GLint bsize;

    // cleanup in case of resize
    if (pixels != NULL)
    {
        free(pixels); pixels = NULL;

        cutilSafeCall(cudaGLUnregisterBufferObject(pbo_buffer));

        glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
        glDeleteBuffers(1, &pbo_buffer);
        glDeleteTextures(1, &texid);
        deleteTexture();
    }

    imWidth = width, imHeight = height;
    g_Bpp = 1;
    pixels = (unsigned int*)calloc(1, wWidth * wHeight * g_Bpp * sizeof(unsigned int));
    setupTexture(imWidth, imHeight, NULL, g_Bpp);

    {
        // use OpenGL Path
        glGenBuffers(1, &pbo_buffer);
        glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo_buffer); 
        glBufferData(GL_PIXEL_UNPACK_BUFFER, 
                        g_Bpp * sizeof(Pixel) * imWidth * imHeight, 
                        pixels, GL_STREAM_DRAW);  

        glGetBufferParameteriv(GL_PIXEL_UNPACK_BUFFER, GL_BUFFER_SIZE, &bsize); 
        if ((GLuint)bsize != (g_Bpp * sizeof(Pixel) * imWidth * imHeight)) {
            printf("Buffer object (%d) has incorrect size (%d).\n", (unsigned)pbo_buffer, (unsigned)bsize);
            cudaThreadExit();
            exit(-1);
        }

        glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
        cutilSafeCall(cudaGLRegisterBufferObject(pbo_buffer));

        glGenTextures(1, &texid);
        glBindTexture(GL_TEXTURE_2D, texid);
//        glTexImage2D(GL_TEXTURE_2D, 0, ((g_Bpp==1) ? GL_LUMINANCE : GL_BGRA), 
//                    imWidth, imHeight,  0, GL_LUMINANCE, GL_UNSIGNED_INT, NULL);
        glTexImage2D(GL_TEXTURE_2D, 0, 4, 
                    imWidth, imHeight,  0, GL_BGRA, GL_UNSIGNED_BYTE, NULL);
        glBindTexture(GL_TEXTURE_2D, 0);

        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
        glPixelStorei(GL_PACK_ALIGNMENT, 1);
    }
}

void printHelp()
{
    printf("\nUsage: Buddhabrot <options>\n");
    printf("\t\t-fbo    (render using FBO display path)\n");
    bQuit = true;
}


//Download wglext.h from http://oss.sgi.com/projects/ogl-sample/sdk.html
#define WGL_NV_gpu_affinity
#include <gl\wglext.h>
#undef WGL_NV_gpu_affinity

PFNWGLSWAPINTERVALEXTPROC wglSwapIntervalEXT = NULL;
PFNWGLGETSWAPINTERVALEXTPROC wglGetSwapIntervalEXT = NULL;


//from NeHe Lesson 45
bool IsExtensionSupported( char* szTargetExtension )
{
    const unsigned char *pszExtensions = NULL;
    const unsigned char *pszStart;
    unsigned char *pszWhere, *pszTerminator;

    // Extension names should not have spaces
    pszWhere = (unsigned char *) strchr( szTargetExtension, ' ' );
    if( pszWhere || *szTargetExtension == '\0' )
        return false;

    // Get Extensions String
    pszExtensions = glGetString( GL_EXTENSIONS );

    // Search The Extensions String For An Exact Copy
    pszStart = pszExtensions;
    for(;;)
    {
        pszWhere = (unsigned char *) strstr( (const char *) pszStart, szTargetExtension );
        if( !pszWhere )
            break;
        pszTerminator = pszWhere + strlen( szTargetExtension );
        if( pszWhere == pszStart || *( pszWhere - 1 ) == ' ' )
            if( *pszTerminator == ' ' || *pszTerminator == '\0' )
                return true;
        pszStart = pszTerminator;
    }
    return false;
}

bool InitVSync()
{
   if (IsExtensionSupported("WGL_EXT_swap_control"))
   {
      wglSwapIntervalEXT = (PFNWGLSWAPINTERVALEXTPROC)
          wglGetProcAddress("wglSwapIntervalEXT");
      wglGetSwapIntervalEXT = (PFNWGLGETSWAPINTERVALEXTPROC)
          wglGetProcAddress("wglGetSwapIntervalEXT");
      return true;
   }
   return false;// WGL_EXT_swap_control not supported
}
bool SetVSync(bool VSync)
{
    if(wglSwapIntervalEXT==NULL) return false;
    wglSwapIntervalEXT(VSync);
    return true;
}
bool GetVSync(bool* VSync)
{
    if(wglGetSwapIntervalEXT==NULL) return false;//VSynce value is not valid...
    *VSync = wglGetSwapIntervalEXT()!=0;
    return true;
}

void initGL(int argc, char** argv)
{
    glutInit( &argc, argv);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
    glutInitWindowSize(wWidth, wHeight);
    glutCreateWindow("CUDA Buddhabrot");
    glutReshapeWindow(wWidth, wHeight);

    glewInit();

    if (g_bFBODisplay) {
        if (!glewIsSupported( "GL_VERSION_2_0 GL_EXT_framebuffer_object" )) {
            fprintf(stderr, "Error: failed to get minimal extensions for demo\n");
            fprintf(stderr, "This sample requires:\n");
            fprintf(stderr, "  OpenGL version 2.0\n");
            fprintf(stderr, "  GL_EXT_framebuffer_object\n");
            cudaThreadExit();
            exit(-1);
        }
    } else {
        if (!glewIsSupported( "GL_VERSION_1_4 GL_ARB_pixel_buffer_object" )) {
            fprintf(stderr, "Error: failed to get minimal extensions for demo\n");
            fprintf(stderr, "This sample requires:\n");
            fprintf(stderr, "  OpenGL version 1.4\n");
            fprintf(stderr, "  GL_ARB_pixel_buffer_object\n");
            cudaThreadExit();
            exit(-1);
        }
    }

    InitVSync();
    SetVSync(false);
}

int main(int argc, char** argv) 
{
    if (argc > 1) {
        cutGetCmdLineArgumenti(argc, (const char **)argv, "width", &wWidth);
        cutGetCmdLineArgumenti(argc, (const char **)argv, "height", &wHeight);
        cutGetCmdLineArgumenti(argc, (const char **)argv, "maxR", &maxR);
        cutGetCmdLineArgumenti(argc, (const char **)argv, "maxG", &maxG);
        cutGetCmdLineArgumenti(argc, (const char **)argv, "maxB", &maxB);
        cutGetCmdLineArgumenti(argc, (const char **)argv, "colormode", &colormode);

        if (cutCheckCmdLineFlag(argc, (const char **)argv, "help")) {
            printHelp();
        }
        if (cutCheckCmdLineFlag(argc, (const char **)argv, "fbo")) {
            g_bFBODisplay = true;
        }
    }

    // First initialize OpenGL context, so we can properly set the GL for CUDA.
    // This is necessary in order to achieve optimal performance with OpenGL/CUDA interop.
    initGL( argc, argv );

    // use command-line specified CUDA device, otherwise use device with highest Gflops/s
    if( cutCheckCmdLineFlag(argc, (const char**)argv, "device") ) {
        cutilGLDeviceInit(argc, argv);
    } else {
        cudaGLSetGLDevice (cutGetMaxGflopsDeviceId() );
    }

    int device;
    struct cudaDeviceProp prop;
    cudaGetDevice( &device );
    cudaGetDeviceProperties( &prop, device );
    if( !strncmp( "Tesla", prop.name, 5 ) ){
        printf("This sample needs a card capable of OpenGL and display.\n");
        printf("Please choose a different device with the -device=x argument.\n");
        cudaThreadExit();
        cutilExit(argc, argv);
    }

    cutilCheckError(cutCreateTimer(&timer));
    cutilCheckError(cutResetTimer(timer));  
 
    glutDisplayFunc(display);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutKeyboardFunc(keyboard);
    glutReshapeFunc(reshape);
    glutIdleFunc(idle);

    initializeData( wWidth, wHeight );

    // If code is not printing the Usage, then we execute this path.
    if (!bQuit) {
        printf("Use the '-' and '=' keys to change the exposure.\n");
        printf("Use the '<' and '>' keys to change the brightness offset.\n");
        printf("'f' key toggles intensity driven denoising filter.\n");
        printf("LMB+drag selects a zoom rectangle, RMB outside clears it.\n");
        printf("Hold Shift to remove fixed aspect ratio during zooming.\n");
        printf("Hold Ctrl to select rectangle around given image center.\n");
        printf("'z' or RMB inside selected region zooms into selection.\n");
        printf("'c' clears the image and renders from scratch.\n");
        printf("Space Bar to reset the zoom to full view.\n");
        fflush(stdout);
        atexit(cleanup); 
        glutMainLoop();
    }

    cudaThreadExit();
    cutilExit(argc, argv);
}
