/*
 * Copyright 2010 Christian Buchner
 */

#include <stdio.h>
#include <stdlib.h>
#include <cutil_inline.h>

// using my custom complex data type here
#include "cudacomplex.h"

#include "Buddhabrot_kernels.h"

// complex circle radius 2
float zrmin = 0.0f, zrmax = 0.0f;
float zimin = 0.0f, zimax = 0.0f;

//Use weak contenders to plot orbits on neighbor threads
#define USEWEAKCONTENDERS 1

//Pass on valid orbit seeds from one thread to the next
#define CROSSPOLINATE 1

// master flag for METROPOLIS HASTINGS algorithm (allows fast zoomed renders)
#define METROPOLIS 1

// Metric 1: original
// Metric 2: weaker influence of orbit length
// Metric 3: no influence of orbit length
#define METRIC 1

// Tile width
#define TW 64

// height of stack for valid random number starting positions
#define STACK 8

// fast image filter variant
#define FAST_NLM 1

// Texture reference for reading image
texture<uchar4, 2, cudaReadModeNormalizedFloat> tex;
static cudaArray *array = NULL;

#define MAX(a,b) ((a)>(b)?(a):(b))

// assuming the rand() function only returns 15 bits of randomness (Visual C++)
__host__ int rand32()
{
   unsigned int foo = rand() ^ (rand() << 15) ^ ((rand() << 30) & 0x3);
   return foo;
}

// draw a random number [0;1)
__host__ __device__ float sfrand( int &seed )
{
    float res;
    seed *= 16807; // other full-period numbers: 47271 or 69621
    *((unsigned int *) &res) = ( ((unsigned int)seed)>>9 ) | 0x3f800000;  // range: 1..2
    return( res - 1.0f );  // range: 0..1
}

// two dimensional normal distribution based on box-muller algorithm
__host__ __device__ void boxmuller( float &yr, float &yi, int &seedr, int &seedi )
{
    float xr, xi, w;
    do {
        xr = 2.0f * sfrand(seedr) - 1.0f;
        xi = 2.0f * sfrand(seedi) - 1.0f;
        w = xr * xr + xi * xi;
    } while ( w >= 1.0f );

    w = sqrt( (-2.0f * logf( w ) ) / w );
    yr = xr * w;
    yi = xi * w;
}

// A kernel to provide good starting orbits for the Buddhakernel
__global__ void
Tileseedkernel( unsigned int *g_hstack,
                singlecomplex *g_coordsstack,
                uint2 *g_lorbitstack,
                singlecomplex *g_coords,
                unsigned int *g_lorbit,
                int *g_rseed, int *g_iseed,
                int w, int h, int th,
                float zrmin, float zrmax, float zimin, float zimax,
                float Rrmin, float Rrmax, float Rimin, float Rimax,
                float isigma, unsigned int maxiter )
{
    float Rr = (Rrmax + Rrmin)/2;
    float Ri = (Rimax + Rimin)/2;
    float Ra = (Rrmax - Rrmin)/2;
    float Rb = (Rimax - Rimin)/2;

    __shared__ int s_seedr[256]; // seeds for real axis
    __shared__ int s_seedi[256]; // seeds for imaginary axis

    int threadoffset = threadIdx.x + blockDim.x * (gridDim.x * blockIdx.y + blockIdx.x);
    int nthreads     = blockDim.x * (gridDim.x * gridDim.y);

    // every thread reads its unique seed
    s_seedr[threadIdx.x] = g_rseed[threadoffset];
    s_seedi[threadIdx.x] = g_iseed[threadoffset];

    unsigned int ssize = g_hstack[threadoffset];

    // perform up to 3 random draws per thread and put all "good" random numbers on the stack
    // these random draws are starting locations for orbits that are within a circle of radius 2
    // and are not within the main cardioid of the Mandelbrot set as well as the four largest bulbs.
    for (int draws = 0; ssize < STACK && draws < 3; draws++)
    {
        float zr, zi;
        unsigned int good = false;

        // generate complex random number (real and imaginary) and scale to target window
        zr = sfrand(s_seedr[threadIdx.x]) * (Rrmax-Rrmin) + Rrmin;
        zi = sfrand(s_seedi[threadIdx.x]) * (Rimax-Rimin) + Rimin;

        // circle (radius 2) test, escape criterion
        if (zr*zr + zi*zi < 4.0f)
        {
            // random generator ellipse test
            if (((zr-Rr)*(zr-Rr)/(Ra*Ra) + (zi-Ri)*(zi-Ri)/(Rb*Rb)) < 1)
            {
                // period-2 bulb test
                if ((((zr+1.0f)*(zr+1.0f)) + zi*zi) >= 0.0625f)
                {
                    // cardioid test
                    float q = (zr-0.25f)*(zr-0.25f) + zi*zi;
                    if (q*(q+(zr-0.25f)) >= 0.25f*zi*zi)
                    {
                        // test for the smaller bulb left of the period-2 bulb
                        if ((((zr+1.309f)*(zr+1.309f)) + zi*zi) >= 0.00345f)
                            // check for the smaller bulbs on top and bottom of the brot
                            if ((((zr+0.125f)*(zr+0.125f)) + (zi-0.744f)*(zi-0.744f)) >= 0.0088f)
                            if ((((zr+0.125f)*(zr+0.125f)) + (zi+0.744f)*(zi+0.744f)) >= 0.0088f)
                                good = true;
                    }
                }
            }
        }

        // perform another test whether or not the orbit affects the visible region
        if (good)
        {
            // iteration loop
            singlecomplex c = make_singlecomplex(zr, zi);
            singlecomplex z = make_singlecomplex(0.0f,0.0f);

            good = 0; // count "good" pixels, i.e. such in the viewable area
            unsigned int iter = 0;
            do
            {
                z = z * z + c;

                // record if we have any pixels within the visible region
                good += (z.real() >= zrmin && z.real() < zrmax &&
                         z.imag() >= zimin && z.imag() < zimax );

            } while (++iter < maxiter && z.abs2() < 4);

            // reject points that may be in the M-Set
            if (iter == maxiter) good = iter = 0;

            // put all "good" starting locations on the stack
            if (good)
            {
                g_coordsstack[nthreads * ssize + threadoffset] = c;
                g_lorbitstack[nthreads * ssize + threadoffset] = make_uint2(iter, good);
                ++ssize;
            }
        }
    }

    // use default values in case the coordinate stack is empty
    singlecomplex c = make_singlecomplex(0.0f, 0.0f);
    unsigned int iter = 0, good = 0;

#if !METROPOLIS

    //
    // Uniform random sampling using only valid random positions from the stack
    //
    
    if (ssize > 0)
    {
        ssize--;
        c = g_coordsstack[nthreads * ssize + threadoffset];
        uint2 tmp = g_lorbitstack[nthreads * ssize + threadoffset];
        iter = tmp.x;
        good = tmp.y;
    }

    // pass the coordinates and orbit length to the Buddhakernel
    g_coords[threadoffset] = c;
    g_lorbit[threadoffset] = iter;

#else

    //
    // Nonuniform sampling using a method similar to Metropolis-Hastings
    //

    unsigned int goodA, goodB, goodC;
    singlecomplex cA,cB,cC,zA,zB,zC;
    unsigned int iterA,iterB,iterC;
    float pA,pB,pC;

    goodA = goodB = goodC = 0;
    cA=cB=cC=zA=zB=zC=make_singlecomplex(0.0f, 0.0f);
    iterA=iterB=iterC=0;
    pA=pB=pC=0;

    // when available, use the top element from the stack as a starting point
    // for three mutations A,B,C
    if (ssize > 0)
    {
        c = g_coordsstack[nthreads * (ssize-1) + threadoffset];
        uint2 tmp = g_lorbitstack[nthreads * (ssize-1) + threadoffset];
        iter = tmp.x;
        good = tmp.y;

        // find three contenders A,B,C which are a mutations of c, plus
        // a 2D gaussian distribution with variance of 1/10th of the visible window
        float d_r, d_i; 

        boxmuller(d_r, d_i, s_seedr[threadIdx.x], s_seedi[threadIdx.x]);
        cA = make_singlecomplex(c.real() + d_r * (zrmax-zrmin)/isigma,
                                c.imag() + d_i * (zimax-zimin)/isigma);

        boxmuller(d_r, d_i, s_seedr[threadIdx.x], s_seedi[threadIdx.x]);
        cB = make_singlecomplex(c.real() + d_r * (zrmax-zrmin)/isigma,
                                c.imag() + d_i * (zimax-zimin)/isigma);

        boxmuller(d_r, d_i, s_seedr[threadIdx.x], s_seedi[threadIdx.x]);
        cC = make_singlecomplex(c.real() + d_r * (zrmax-zrmin)/isigma,
                                c.imag() + d_i * (zimax-zimin)/isigma);

        // compact iteration loop for all three contenders
        bool cont;
        do {
            cont = false;
            if (iterA < maxiter && zA.abs2() < 4) {
                iterA++; zA=zA*zA+cA;
                goodA += (zA.real()>=zrmin && zA.real()<zrmax && zA.imag()>=zimin && zA.imag()<zimax );
                cont=true;
            }
            if (iterB < maxiter && zB.abs2() < 4) {
                iterB++; zB=zB*zB+cB;
                goodB += (zB.real()>=zrmin && zB.real()<zrmax && zB.imag()>=zimin && zB.imag()<zimax );
                cont=true;
            }
            if (iterC < maxiter && zC.abs2() < 4) {
                iterC++; zC=zC*zC+cC;
                goodC += (zC.real()>=zrmin && zC.real()<zrmax && zC.imag()>=zimin && zC.imag()<zimax );
                cont=true;
            }
        } while(cont);

        // reject points that may be in the M-Set
        if (iterA == maxiter) goodA = iterA = 0;
        if (iterB == maxiter) goodB = iterB = 0;
        if (iterC == maxiter) goodC = iterC = 0;

        // three variants of weighting metrics, pick one.

#if METRIC == 1
        pA = ((float)goodA * iterA * iterA) / ((float)good * iter * iter);
        pB = ((float)goodB * iterB * iterB) / ((float)good * iter * iter);
        pC = ((float)goodC * iterC * iterC) / ((float)good * iter * iter);
#else
#if METRIC == 2
        pA = ((float)goodA / (float)good) * ((float)iterA / (float)iter);
        pB = ((float)goodB / (float)good) * ((float)iterB / (float)iter);
        pC = ((float)goodC / (float)good) * ((float)iterC / (float)iter);
#else
#if METRIC == 3
        float pA = (float)goodA / (float)good;
        float pB = (float)goodB / (float)good;
        float pC = (float)goodC / (float)good;
#endif
#endif
#endif

        float pX = 1.0f - MAX(pA, MAX(pB, pC));
        if (pX < 0.0f) pX = 0.0f;

        float pSum = pX + pA + pB + pC;
        float R = pSum * sfrand(s_seedr[threadIdx.x]);

        // this case means we decided to not pick a mutation over the current seed
        if (R < pX)
        {
            // with a low probability throw the current element from the stack
            // to allow for some fluctuation of seeds
            if (sfrand(s_seedi[threadIdx.x]) < (float)ssize/10000) ssize--;

#if USEWEAKCONTENDERS
            // plot the orbit with the largest valid contribution, but do not
            // replace the seed value on the stack. So the mutation proceeds
            // in most optimal way, but we still have something useful to plot.
            // from any of the "failed candidates" we pick the best one
                 if (goodA && pA >= MAX(pB, pC)) { c = cA; iter = iterA; }
            else if (goodB && pB >= MAX(pA, pC)) { c = cB; iter = iterB; }
            else if (goodC && pC >= MAX(pA, pB)) { c = cC; iter = iterC; }
            // no usable contender found
            else
#endif
                iter = 0;
        }
        else if (R < pX + pA)
        {
            // contender A replaces topmost element on stack
            c = cA; iter = iterA;
            g_coordsstack[nthreads * (ssize-1) + threadoffset] = c;
            g_lorbitstack[nthreads * (ssize-1) + threadoffset] = make_uint2(iter, goodA);
        }
        else if (R < pX + pA + pB)
        {
            // contender B replaces topmost element on stack
            c = cB; iter = iterB;
            g_coordsstack[nthreads * (ssize-1) + threadoffset] = c;
            g_lorbitstack[nthreads * (ssize-1) + threadoffset] = make_uint2(iter, goodB);
        }
        else
        {
            // contender C replaces topmost element on stack
            c = cC; iter = iterC;
            g_coordsstack[nthreads * (ssize-1) + threadoffset] = c;
            g_lorbitstack[nthreads * (ssize-1) + threadoffset] = make_uint2(iter, goodC);
        }
    }

#if USEWEAKCONTENDERS && CROSSPOLINATE
    // pass on one of our unused contenders to the direct neighbor threads if possible
    // from the "failed candidates", pick the best one - just not the same one
    // that the current thread will already be rendering already.

    __shared__ singlecomplex s_c[256];
    __shared__ unsigned int  s_l[256];

    s_c[threadIdx.x] = c;
    s_l[threadIdx.x] = iter;
    __syncthreads();

    bool flag = false;
    if (threadIdx.x > 0 && s_l[threadIdx.x-1] == 0)
    {
        if (c == cA) {
            if      (goodB && pB >= pC) { flag = true; s_c[threadIdx.x-1] = cB; s_l[threadIdx.x-1] = iterB; }
            else if (goodC && pC >= pB) { flag = true; s_c[threadIdx.x-1] = cC; s_l[threadIdx.x-1] = iterC; }
        }
        if (c == cB) {
            if      (goodA && pA >= pC) { flag = true; s_c[threadIdx.x-1] = cA; s_l[threadIdx.x-1] = iterA; }
            else if (goodC && pC >= pA) { flag = true; s_c[threadIdx.x-1] = cC; s_l[threadIdx.x-1] = iterC; }
        }
        if (c == cC) {
            if      (goodA && pA >= pB) { flag = true; s_c[threadIdx.x-1] = cA; s_l[threadIdx.x-1] = iterA; }
            else if (goodB && pB >= pC) { flag = true; s_c[threadIdx.x-1] = cB; s_l[threadIdx.x-1] = iterB; }
        }
    }
    __syncthreads();
    if (!flag && threadIdx.x < blockDim.x-1 && s_l[threadIdx.x+1] == 0)
    {
        if (c == cA) {
            if      (goodB && pB >= pC) { flag = true; s_c[threadIdx.x+1] = cB; s_l[threadIdx.x+1] = iterB; }
            else if (goodC && pC >= pB) { flag = true; s_c[threadIdx.x+1] = cC; s_l[threadIdx.x+1] = iterC; }
        }
        if (c == cB) {
            if      (goodA && pA >= pC) { flag = true; s_c[threadIdx.x+1] = cA; s_l[threadIdx.x+1] = iterA; }
            else if (goodC && pC >= pA) { flag = true; s_c[threadIdx.x+1] = cC; s_l[threadIdx.x+1] = iterC; }
        }
        if (c == cC) {
            if      (goodA && pA >= pB) { flag = true; s_c[threadIdx.x+1] = cA; s_l[threadIdx.x+1] = iterA; }
            else if (goodB && pB >= pC) { flag = true; s_c[threadIdx.x+1] = cB; s_l[threadIdx.x+1] = iterB; }
        }
    }
    __syncthreads();
    g_coords[threadoffset] = s_c[threadIdx.x];
    g_lorbit[threadoffset] = s_l[threadIdx.x];
#else
    g_coords[threadoffset] = c;
    g_lorbit[threadoffset] = iter;
#endif  // #if USEWEAKCONTENDERS

#endif  // #if !METROPOLIS

    // store the resulting stack size
    g_hstack[threadoffset] = ssize;

#if CROSSPOLINATE
    // Pass on some valid random coordinates between threads to jump-start the initial seeding
    // process. We even allow passing data across block boundaries. The lack of inter-block
    // synchronization may be a little dangerous, but hey...
    __syncthreads();
    if (ssize == 0 && threadoffset > 0)
    {
        // clone the predecessor thread ID's stacked seed
        if (g_hstack[threadoffset-1] > 0)
        {
            g_coordsstack[threadoffset] = g_coordsstack[threadoffset-1];
            g_lorbitstack[threadoffset] = g_lorbitstack[threadoffset-1];
            g_hstack[threadoffset] = ++ssize;
        }
    }
    __syncthreads();
    if (ssize == 0 && threadoffset < nthreads-1)
    {
        // clone the successor thread ID's stacked seed
        if (g_hstack[threadoffset+1] > 0)
        {
            g_coordsstack[threadoffset] = g_coordsstack[threadoffset+1];
            g_lorbitstack[threadoffset] = g_lorbitstack[threadoffset+1];
            g_hstack[threadoffset] = ++ssize;
        }
    }
    __syncthreads();
#endif

    // save the thread-unique random seed
    g_rseed[threadoffset] = s_seedr[threadIdx.x];
    g_iseed[threadoffset] = s_seedi[threadIdx.x];
}

#ifdef __DEVICE_EMULATION__
#define EMUSYNC __syncthreads()
#else
#define EMUSYNC
#endif


// A kernel to count the number of "active" threads
__global__ void 
Analysiskernel( Pixel *pAccumBuffer,
                singlecomplex *g_coords,
                unsigned int *g_lorbit,
                unsigned int Pitch,  int w, int h, int th, bool clear,
                float zrmin, float zrmax, float zimin, float zimax,
                unsigned int *g_result, unsigned int maxiter)
{
    int tid = threadIdx.x;
    int block = gridDim.x * blockIdx.y + blockIdx.x;

    __shared__ unsigned int sdata[256];

    // iteration loop
    unsigned int n = g_lorbit[threadIdx.x + blockDim.x * block];

    // set a flag if each thread has work
    sdata[threadIdx.x] = ( n> 0 && n != maxiter ) ? 1 : 0;
    __syncthreads();

#define blockSize 256
    if (blockSize >= 256) { if (tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
    if (blockSize >= 128) { if (tid <  64) { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }
    
#ifndef __DEVICE_EMULATION__
    if (tid < 32)
#endif
    {
        if (blockSize >=  64) { sdata[tid] += sdata[tid + 32]; EMUSYNC; }
        if (blockSize >=  32) { sdata[tid] += sdata[tid + 16]; EMUSYNC; }
        if (blockSize >=  16) { sdata[tid] += sdata[tid +  8]; EMUSYNC; }
        if (blockSize >=   8) { sdata[tid] += sdata[tid +  4]; EMUSYNC; }
        if (blockSize >=   4) { sdata[tid] += sdata[tid +  2]; EMUSYNC; }
        if (blockSize >=   2) { sdata[tid] += sdata[tid +  1]; EMUSYNC; }
    }

    if (threadIdx.x == 0)
        atomicAdd(g_result, sdata[0]);
}


__device__ void    WavelengthToRGB(float Wavelength , float &r , float &g , float &b)
{
    // After Bruton's (astro@tamu.edu) FORTRAN code

    float sss, gamma;

    gamma = 0.45f;    // Bruton had 0.80 - Poynton says 0.45

    if (Wavelength < 380) Wavelength = 380;
    if (Wavelength > 780) Wavelength = 780;

    if ((Wavelength >= 380) && (Wavelength < 440))
    {
        r = -1.0f * (Wavelength - 440.0f) / (440.0f - 380.0f);
        g = 0;
        b = 1.0f;
    }
    else if ((Wavelength >= 440) && (Wavelength < 490))
    {
        r = 0;
        g = (Wavelength - 440.0f) / (490.0f - 440.0f);
        b = 1.0f;
    }
    else if ((Wavelength >= 490) && (Wavelength < 510))
    {
        r = 0;
        g = 1.0f;
        b = -1.0f * (Wavelength - 510.0f) / (510.0f - 490.0f);
    }
    else if ((Wavelength >= 510) && (Wavelength < 580))
    {
        r = (Wavelength - 510.0f) / (580.0f - 510.0f);
        g = 1.0f;
        b = 0;
    }
    else if ((Wavelength >= 580) && (Wavelength < 645))
    {
        r = 1.0f;
        g = -1.0f * (Wavelength - 645.0f) / (645.0f - 580.0f);
        b = 0;
    }
    else
    {
        r = 1.0f;
        g = 0;
        b = 0;
    }

    if (Wavelength > 700)
    {
        sss = 0.3f + 0.7f * (780.0f - Wavelength) / (780.0f - 700.0f);
    }
    else if (Wavelength < 420)
    {
        sss = 0.3f + 0.7f * (Wavelength - 380.0f) / (420.0f - 380.0f);
    }
    else
    {
        sss = 1.0f;
    }

    r = pow(sss * r , gamma);
    g = pow(sss * g , gamma);
    b = pow(sss * b , gamma);

    return;
}

// A kernel to plot orbits into an accumulation buffer
// This is currently monochrome. Writes are not atomic, collisions can occur but should be rare.
// The tiled rendering approach was supposed to render to small tiles in shared memory, but was
// given up because achieving a fast performance would have been very difficult.
__global__ void 
Buddhakernel( Pixel *pAccumBuffer,
              singlecomplex *g_coords,
              unsigned int *g_lorbit,
              int *g_rseed, int *g_iseed,
              unsigned int Pitch,  int w, int h, int th, bool clear,
              float zrmin, float zrmax, float zimin, float zimax,
              unsigned int maxiter, int maxR, int maxG, int maxB,
              int colormode)
{
//    int threadoffset = threadIdx.x + blockDim.x * (gridDim.x * blockIdx.y + blockIdx.x);

    int block = gridDim.x * blockIdx.y + blockIdx.x;

    __shared__ int s_seedr[256]; // seeds for real axis
    __shared__ int s_seedi[256]; // seeds for imaginary axis

    int threadoffset = threadIdx.x + blockDim.x * block;

    if (colormode != 0)
    {
        // every thread reads its unique seed
        s_seedr[threadIdx.x] = g_rseed[threadoffset];
        s_seedi[threadIdx.x] = g_iseed[threadoffset];
    }

    // number of tile's rows that can be handled within this thread block
    int nrows = blockDim.x  / TW;

    // thread specific location within tile
    int yoff  = threadIdx.x / TW;
    int xoff  = threadIdx.x % TW;

    // (x,y) absolute top left tile coordinates in pixels (inclusive)
    // (X,Y) absolute bottom right tile coordinates in pixels (exclusive)
    int tile_xpos =  blockIdx.x    * TW;
    int tile_ypos =  blockIdx.y    * th;

    if (clear)
    {
        unsigned int *out = pAccumBuffer + Pitch * (tile_ypos + yoff) + tile_xpos + xoff;
        for (int yrow=0; yrow < th; yrow+=nrows) {
            if (yrow + yoff < th && tile_ypos + yrow + yoff < h)
                *out = 0;
            out += nrows * Pitch;
        }
    }

    // generate a Mandelbrot orbit at the given complex starting coordinates
    singlecomplex c = g_coords[threadIdx.x + blockDim.x * block];
    singlecomplex z = make_singlecomplex(0.0f,0.0f);

    // iteration loop
    unsigned int n = g_lorbit[threadIdx.x + blockDim.x * block];
    if (n> 0 && n != maxiter && true)
    {

        unsigned int add;

        float r,g,b;

        if (colormode == 0)       // Traditional coloring
        {
            // depending on orbit length, choose to which color
            // components to add for this orbit.
            add = 0;
            if (n <= maxR) add |= 0x100000;  // pow(2,20): 12 bits
            if (n <= maxG) add |= 0x000400;  // pow(2,10): 10 bits
            if (n <= maxB) add |= 0x000001;  // pow(2, 1): 10 bits
        }
        else if (colormode == 1)  // Physics wavelength based coloring
        {
            WavelengthToRGB((float)maxR+(float)n*400.0f/maxG, r,g,b);
        }

        unsigned int iter = 0;
        do
        {
            z = z * z + c;

            ++iter;
            if (z.abs2() < 4 && true)  // add some extra conditions to prevent or abort iterating if you like
            {
                if (true) // add some extra conditions to restrict plotting if you like
                {
                    int xabs = (z.real() - zrmin) / (zrmax - zrmin) * (w-1) + 0.5f;
                    int yabs = (z.imag() - zimin) / (zimax - zimin) * (h-1) + 0.5f;
                    if (xabs >= 0 && xabs < w &&
                        yabs >= 0 && yabs < h)
                    {
                        if (colormode == 1)
                        {
                            add = 0;
                            if (sfrand(s_seedr[threadIdx.x]) < r) add |= 0x100000;  // pow(2,20): 12 bits
                            if (sfrand(s_seedr[threadIdx.x]) < g) add |= 0x000400;  // pow(2,10): 10 bits
                            if (sfrand(s_seedr[threadIdx.x]) < b) add |= 0x000001;  // pow(2, 1): 10 bits
                        }

                        unsigned int *out = pAccumBuffer + Pitch * (yabs) + xabs;
                        *out += add;
                    }
                }
            } else break;
        } while (iter < maxiter);
    }

    if (colormode != 0)
    {
        // save the thread-unique random seed
        g_rseed[threadoffset] = s_seedr[threadIdx.x];
        g_iseed[threadoffset] = s_seedi[threadIdx.x];
    }
}

////////////////////////////////////////////////////////////////////////////////
// Filter configuration
////////////////////////////////////////////////////////////////////////////////
#define KNN_WINDOW_RADIUS   3
#define NLM_WINDOW_RADIUS   3
#define NLM_BLOCK_RADIUS    3
#define KNN_WINDOW_AREA     ( (2 * KNN_WINDOW_RADIUS + 1) * (2 * KNN_WINDOW_RADIUS + 1) )
#define NLM_WINDOW_AREA     ( (2 * NLM_WINDOW_RADIUS + 1) * (2 * NLM_WINDOW_RADIUS + 1) )
#define INV_KNN_WINDOW_AREA ( 1.0f / (float)KNN_WINDOW_AREA )
#define INV_NLM_WINDOW_AREA ( 1.0f / (float)NLM_WINDOW_AREA )

#define KNN_WEIGHT_THRESHOLD    0.02f
#define KNN_LERP_THRESHOLD      0.79f
#define NLM_WEIGHT_THRESHOLD    0.10f
#define NLM_LERP_THRESHOLD      0.10f

#define BLOCKDIM_X 8
#define BLOCKDIM_Y 8

int iDivUp(int a, int b){
    return ((a % b) != 0) ? (a / b + 1) : (a / b);
}
__device__ float lerpf(float a, float b, float c){
    return a + (b - a) * c;
}

__device__ float vecLen(float4 a, float4 b){
    return (
        (b.x - a.x) * (b.x - a.x) +
        (b.y - a.y) * (b.y - a.y) +
        (b.z - a.z) * (b.z - a.z)
    );
}

__device__ float vecLen(float a, float b){
    return (b - a) * (b - a) * 3;
}

#define M_PI 3.14159265f
__device__ float transition(float x, float threshold) {
    return 0.5f + atan(-(x-threshold)/(M_PI*3.f))/M_PI;
}

////////////////////////////////////////////////////////////////////////////////
//   0 1 2 3 4 5 6 7 
// 0 + . . . . . . . 
// 1 . . . . . . . .
// 2 . . . . . . . . 
// 3 . . . * . . . .
// 4 . . . . . . . .
// 5 . . . . . . . .
// 6 . . . . . . . .
// 7 . . . . . . . .
//
// * - Base point for every thread, + - pixel around which ColorDistance is computed
// The idea behind this method:
// - Every thread in a 8x8 block computes just one ColorDistance
// - It is saved in the weights array that is shared accross the threads
// - Threads are synced
// - For every pixel inside the block weights are considered to be constants
////////////////////////////////////////////////////////////////////////////////
__global__ void NLM2(
    Pixel *dst,
    int imageW,
    int imageH,
    float Noise,
    float lerpC
){
#if FAST_NLM
    //Weights cache
    __shared__ float fWeights[BLOCKDIM_X * BLOCKDIM_Y];

    const int   ix = blockDim.x * blockIdx.x + threadIdx.x;
    const int   iy = blockDim.y * blockIdx.y + threadIdx.y;
    //Add half of a texel to always address exact texel centers
    const float  x = (float)ix  + 0.5f;
    const float  y = (float)iy  + 0.5f;
    const float cx = blockDim.x * blockIdx.x + NLM_WINDOW_RADIUS + 0.5f;
    const float cy = blockDim.x * blockIdx.y + NLM_WINDOW_RADIUS + 0.5f;

    if(ix < imageW && iy < imageH){
        //Find color distance from current texel to the center of NLM window
        float weight = 0;
        for(float n = -NLM_BLOCK_RADIUS; n <= NLM_BLOCK_RADIUS; n++)
            for(float m = -NLM_BLOCK_RADIUS; m <= NLM_BLOCK_RADIUS; m++)
                weight += vecLen(
                    tex2D(tex, cx + m, cy + n),
                    make_float4(0.0f,0.0f,0.0f,0.0f)  // tex2D(tex,  x + m,  y + n)
                );

        //Geometric distance from current texel to the center of NLM window
        float dist = 
            (threadIdx.x - NLM_WINDOW_RADIUS) * (threadIdx.x - NLM_WINDOW_RADIUS) + 
            (threadIdx.y - NLM_WINDOW_RADIUS) * (threadIdx.y - NLM_WINDOW_RADIUS);

        //Derive final weight from color and geometric distance
        weight = __expf( -(weight * Noise + dist * INV_NLM_WINDOW_AREA) );

        //Write the result to shared memory
        fWeights[threadIdx.y * BLOCKDIM_X + threadIdx.x] = weight;
        //Wait until all the weights are ready
        __syncthreads();


        //Normalized counter for the NLM weight threshold
        float fCount = 0;
        //Total sum of pixel weights
        float sumWeights = 0;
        //Result accumulator
        float4 clr = make_float4(0.0f,0.0f,0.0f,0.0f);

        int idx = 0;

        //Cycle through NLM window, surrounding (x, y) texel
        for(float i = -NLM_WINDOW_RADIUS; i <= NLM_WINDOW_RADIUS + 1; i++)
        {
            for(float j = -NLM_WINDOW_RADIUS; j <= NLM_WINDOW_RADIUS + 1; j++)
            {
                //Load precomputed weight
                float weightIJ = fWeights[idx++];

                //Accumulate (x + j, y + i) texel color with computed weight
                float4 clrIJ = tex2D(tex, x + j, y + i);
                clr.x += clrIJ.x * weightIJ;
                clr.y += clrIJ.y * weightIJ;
                clr.z += clrIJ.z * weightIJ;
                clr.w += clrIJ.w * weightIJ;

                //Sum of weights for color normalization to [0..1] range
                sumWeights  += weightIJ;

                //Update weight counter, if NLM weight for current window texel
                //exceeds the weight threshold
                fCount      += (weightIJ > NLM_WEIGHT_THRESHOLD) ? INV_NLM_WINDOW_AREA : 0;
            }
        }

        float lerpQ = transition(sumWeights, NLM_WEIGHT_THRESHOLD * NLM_WINDOW_AREA);

        //Normalize result color by sum of weights
        sumWeights = 1.0f / sumWeights;
        clr.x *= sumWeights;
        clr.y *= sumWeights;
        clr.z *= sumWeights;
        clr.w *= sumWeights;

        //Choose LERP quotent basing on how many texels
        //within the NLM window exceeded the weight threshold
//        float lerpQ = (fCount > NLM_LERP_THRESHOLD) ? lerpC : 1.0f - lerpC;

        //Write final result to global memory
        float4 clr00 = tex2D(tex, x, y);
        clr = make_float4(lerpf(clr.x, clr00.x, lerpQ),
                          lerpf(clr.y, clr00.y, lerpQ),
                          lerpf(clr.z, clr00.z, lerpQ),
                          lerpf(clr.w, clr00.w, lerpQ));

        dst[imageW * iy + ix] = (unsigned int)(clr.x*255.f)|
                                (unsigned int)(clr.y*255.f)<<8|
                                (unsigned int)(clr.z*255.f)<<16|
                                (unsigned int)(clr.w*255.f)<<24;
    }

#else

    const int ix = blockDim.x * blockIdx.x + threadIdx.x;
    const int iy = blockDim.y * blockIdx.y + threadIdx.y;
    //Add half of a texel to always address exact texel centers
    const float x = (float)ix + 0.5f;
    const float y = (float)iy + 0.5f;

    if(ix < imageW && iy < imageH){
        //Normalized counter for the NLM weight threshold
        float fCount = 0;
        //Total sum of pixel weights
        float sumWeights = 0;
        //Result accumulator
        float clr = 0;

        //Cycle through NLM window, surrounding (x, y) texel
        for(float i = -NLM_WINDOW_RADIUS; i <= NLM_WINDOW_RADIUS; i++)
            for(float j = -NLM_WINDOW_RADIUS; j <= NLM_WINDOW_RADIUS; j++)
            {
                //Find color distance from (x, y) to (x + j, y + i)
                float weightIJ = 0;
                for(float n = -NLM_BLOCK_RADIUS; n <= NLM_BLOCK_RADIUS; n++)
                    for(float m = -NLM_BLOCK_RADIUS; m <= NLM_BLOCK_RADIUS; m++)
                        weightIJ += vecLen(
                            tex2D(tex, x + j + m, y + i + n),
                            0.0f //tex2D(tex,     x + m,     y + n)
                        );

                //Derive final weight from color and geometric distance
                weightIJ     = __expf( -(weightIJ * Noise + (i * i + j * j) * INV_NLM_WINDOW_AREA) );

                //Accumulate (x + j, y + i) texel color with computed weight
                float clrIJ = tex2D(tex, x + j, y + i);
                clr       += clrIJ * weightIJ;

                //Sum of weights for color normalization to [0..1] range
                sumWeights  += weightIJ;

                //Update weight counter, if NLM weight for current window texel
                //exceeds the weight threshold
                fCount      += (weightIJ > NLM_WEIGHT_THRESHOLD) ? INV_NLM_WINDOW_AREA : 0;
            }

        float lerpQ = transition(sumWeights, NLM_WEIGHT_THRESHOLD * NLM_WINDOW_AREA);

        //Normalize result color by sum of weights
        sumWeights = 1.0f / sumWeights;
        clr *= sumWeights;

        //Choose LERP quotent basing on how many texels
        //within the NLM window exceeded the weight threshold
//        float lerpQ = (fCount > NLM_LERP_THRESHOLD) ? lerpC : 1.0f - lerpC;

        //Write final result to global memory
        float clr00 = tex2D(tex, x, y);
        clr = lerpf(clr, clr00, lerpQ);
        dst[imageW * iy + ix] = (unsigned int)(clr*4294967295.f);
    }
#endif
}


__global__ void 
HDR2LDRKernel(Pixel *pAccumBuffer,
              Pixel *pOverflowBuffer,
              Pixel *pDisplayBuffer,
              unsigned int Pitch, int w, int h, int th,
              int offset, float scale )
{
    // number of tile's rows that can be handled within this thread block
    int nrows = blockDim.x  / TW;

    // thread specific location within tile
    int yoff  = threadIdx.x / TW;
    int xoff  = threadIdx.x % TW;

    // (x,y) absolute top left tile coordinates in pixels (inclusive)
    int tile_xpos =  blockIdx.x    * TW;
    int tile_ypos =  blockIdx.y    * th;

    // tile[th][TW] is a dynamic shared memory block
    // NOTE: TW must be a constant expression for this to work
    extern __shared__ unsigned int shared[];
    unsigned int (*tile)[TW] = (unsigned int (*)[TW]) shared;

    float foffset = ((offset < 0) ? -1.f : 1.f) * powf(fabs((float)offset), 0.5f);

    // read the tile's contents from global memory and scale accordingly
    unsigned int off  = Pitch * (tile_ypos + yoff) + tile_xpos + xoff;
    unsigned int *acc = pAccumBuffer    + off;
    unsigned int *ovr = pOverflowBuffer + off;

    for (int yrow=0; yrow < th; yrow+=nrows) {
        if (yrow + yoff < th && tile_ypos + yrow + yoff < h)
        {
#if 0
            // GL_LUMINANCE, GL_LUMINANCE, GL_UNSIGNED_INT mapping
            float tmp = (powf((float)*in, 0.5) + foffset)*scale;
            if (tmp < 0) tmp = 0; else if (tmp > 255) tmp = 255;
            tile[yrow+yoff][xoff] = (unsigned int)(tmp*16777216.f);
#else
            // 4, GL_BGRA, GL_UNSIGNED_CHAR grayscale mapping
            unsigned int components = *acc;
            unsigned int overflow   = *ovr;

            // the highest bit of each component is carried over into to overflow
            unsigned int new_overflow = (components&0x80080200) >> 9;
            if (new_overflow != 0)
            {
                components = (components&0x7ff7fdff);
                overflow  += new_overflow;

                *acc = components;
                *ovr = overflow;
            }
            float tmpR = (powf((float)( ((components&0xfff00000)>>20) + ((overflow&0xfff00000) >> 11) ), 0.5) + foffset)*scale;
            float tmpG = (powf((float)( ((components&0x000ffc00)>>10) + ((overflow&0x000ffc00) >>  1) ), 0.5) + foffset)*scale;
            float tmpB = (powf((float)( ((components&0x000003ff)    ) + ((overflow&0x000003ff) <<  9) ), 0.5) + foffset)*scale;
            if (tmpR < 0) tmpR = 0; else if (tmpR > 255) tmpR = 255;
            if (tmpG < 0) tmpG = 0; else if (tmpG > 255) tmpG = 255;
            if (tmpB < 0) tmpB = 0; else if (tmpB > 255) tmpB = 255;
            tile[yrow+yoff][xoff] = (unsigned int)tmpB | (((unsigned int)tmpG)<<8) | (((unsigned int)tmpR)<<16);
#endif
        }
        acc += nrows * Pitch;
        ovr += nrows * Pitch;
    }

    __syncthreads();

    // write back the tile's contents to global memory.
    unsigned int *out = pDisplayBuffer + Pitch * (tile_ypos + yoff) + tile_xpos + xoff;
    for (int yrow=0; yrow < th; yrow+=nrows) {
        if (yrow + yoff < th && tile_ypos + yrow + yoff < h)
            *out = tile[yrow+yoff][xoff];
        out += nrows * Pitch;
    }
}


extern "C" void setupTexture(int iw, int ih, Pixel *data, int Bpp)
{
    cudaChannelFormatDesc desc;

    desc = cudaCreateChannelDesc<uchar4>();

    cutilSafeCall(cudaMallocArray(&array, &desc, iw, ih));
    if (data != NULL)
        cutilSafeCall(cudaMemcpyToArray(array, 0, 0, data, Bpp*sizeof(Pixel)*iw*ih, cudaMemcpyHostToDevice));
}

extern "C" void deleteTexture(void)
{
    cutilSafeCall(cudaFreeArray(array));
    array = NULL;
}

bool init = false;
int initw = 0;
int inith = 0;
int *rseed, *iseed;

Pixel *accumdata;
Pixel *overflowdata;

// per thread orbit starting point
singlecomplex *coords;
unsigned int *lorbit;

// random number stack
unsigned int *hstack;
singlecomplex *coordsstack;
uint2 *lorbitstack;

// atomic counter
unsigned int *d_result;

extern "C" void Zoom(int iw, int ih, int xmin, int xmax, int ymin, int ymax)
{
    float dr = zrmax-zrmin;
    float di = zimax-zimin;
    zrmax = zrmin + dr*(float)xmax / iw;
    zrmin = zrmin + dr*(float)xmin / iw;
    zimax = zimin + di*(float)ymax / ih;
    zimin = zimin + di*(float)ymin / ih;
}

extern "C" void ResetZoom(int iw, int ih)
{
    zrmin = -2.0f*iw/ih; zrmax = 2.0f*iw/ih;
    zimin = -2.0f;       zimax = 2.0f;
}

// Wrapper for the __global__ call that sets up the texture and threads
extern "C" void Buddhabrot(Pixel *displaydata, int iw, int ih, bool clear, int Bpp, int offset, float scale, bool denoise, float isigma, unsigned int maxiter, int maxR, int maxG, int maxB, int colormode) 
{
#if 0
    int test = 1;
    for (int i=0; i < 1000; i++)
    {
        float tmp = sfrand(test);
        fprintf(stderr, "%f\n", tmp);
    }
    exit(1);
#endif

    unsigned int TILEW = TW;
    unsigned int TILEH = 63;
#if TILED
    unsigned int shared = TILEW*TILEH*sizeof(unsigned int);
#else
    unsigned int shared = 0;
#endif

    dim3 grid((iw + TILEW-1)/TILEW, (ih + TILEH-1)/TILEH);

    static bool print = false;
    if (!print) {
        fprintf(stderr, "using a %d x %d grid of %dx%d tiles\n", grid.x, grid.y, TILEW, TILEH);
        print=true;
    }

    unsigned int nThreads = 256; // multiple of TILEW please!!!

    if (zrmin == 0.0f && zrmax == 0.0f && zimin == 0.0f && zimax == 0.0f)
    {
        zrmin = -2.0f*iw/ih; zrmax = 2.0f*iw/ih;
        zimin = -2.0f;       zimax = 2.0f;
    }

    // random generator range
    float Rrmin = -2.0, Rrmax = 2.0;
    float Rimin = -2.0, Rimax = 2.0;

    unsigned int nSeeds = nThreads * grid.x * grid.y;
    
    // window resizing support
    if (init && (initw != iw || inith != ih))
    {
        cutilSafeCall(cudaFree(accumdata));      accumdata = NULL;
        cutilSafeCall(cudaFree(overflowdata));   overflowdata = NULL;
        cutilSafeCall(cudaFree(hstack));         hstack = NULL;
        cutilSafeCall(cudaFree(coordsstack));    coordsstack = NULL;
        cutilSafeCall(cudaFree(lorbitstack));    lorbitstack = NULL;
        cutilSafeCall(cudaFree(coords));         coords = NULL;
        cutilSafeCall(cudaFree(lorbit));         lorbit = NULL;
        cutilSafeCall(cudaFree(rseed));          rseed = NULL;
        cutilSafeCall(cudaFree(iseed));          iseed = NULL;
        cutilSafeCall(cudaFree(d_result));       d_result = NULL;
        init = false;

        float ratio = (float)iw/ih;
        float cx = (zrmin+zrmax)/2;
        float dy = (zimax-zimin);
        zrmin = cx - (ratio * dy/2);
        zrmax = cx + (ratio * dy/2);
    }

    if (!init)
    {
        cutilSafeCall(cudaMalloc( &accumdata,    iw*ih*Bpp    * sizeof(Pixel) ));
        cutilSafeCall(cudaMalloc( &overflowdata, iw*ih*Bpp    * sizeof(Pixel) ));
        cutilSafeCall(cudaMalloc( &hstack,       nSeeds       * sizeof(unsigned int) ));
        cutilSafeCall(cudaMalloc( &coordsstack,  nSeeds*STACK * sizeof(singlecomplex) ));
        cutilSafeCall(cudaMalloc( &lorbitstack,  nSeeds*STACK * sizeof(uint2) ));
        cutilSafeCall(cudaMalloc( &coords,       nSeeds       * sizeof(singlecomplex) ));
        cutilSafeCall(cudaMalloc( &lorbit,       nSeeds       * sizeof(unsigned int) ));
        cutilSafeCall(cudaMalloc( &rseed,        nSeeds       * sizeof(int) ));
        cutilSafeCall(cudaMalloc( &iseed,        nSeeds       * sizeof(int) ));
        cutilSafeCall(cudaMalloc( &d_result,     1            * sizeof(unsigned int) ));

        cutilSafeCall(cudaMemset( accumdata, 0,  iw*ih*Bpp    * sizeof(Pixel) ));
        cutilSafeCall(cudaMemset( overflowdata,0,iw*ih*Bpp    * sizeof(Pixel) ));

        init = true;
        initw = iw;
        inith = ih;
        clear = true;
    }

    if (clear)
    {
        // new seeds
        int *tmp = (int*)calloc(1, nSeeds * sizeof(int));
        for (unsigned int i=0; i < nSeeds; i++) while ((tmp[i] = rand32()) == 0);
        cutilSafeCall(cudaMemcpy( rseed, tmp, nSeeds * sizeof(int), cudaMemcpyHostToDevice ));
        for (unsigned int i=0; i < nSeeds; i++) while ((tmp[i] = rand32()) == 0);
        cutilSafeCall(cudaMemcpy( iseed, tmp, nSeeds * sizeof(int), cudaMemcpyHostToDevice ));
        free(tmp);

        // clear overflow buffer too, because Buddha kernel doesn't do it
        cutilSafeCall(cudaMemset( overflowdata,0,iw*ih*Bpp    * sizeof(Pixel) ));

        // clear random number stack
        cutilSafeCall(cudaMemset( hstack, 0,    nSeeds * sizeof(unsigned int) ));
    }

    Tileseedkernel<<<grid, nThreads, 0>>>( hstack, coordsstack, lorbitstack,
                                           coords, lorbit,
                                           rseed, iseed,
                                           iw, ih, TILEH,
                                           zrmin, zrmax, zimin, zimax,
                                           Rrmin, Rrmax, Rimin, Rimax,
                                           isigma, maxiter );
    cutilCheckMsg("Tileseedkernel execution failed");

    cutilSafeCall(cudaMemset( d_result, 0,    1 * sizeof(unsigned int) ));
    Analysiskernel<<<grid, nThreads, 0>>>( accumdata,
                                           coords, lorbit,
                                           iw, iw, ih, TILEH,
                                           clear,
                                           zrmin, zrmax, zimin, zimax,
                                           d_result, maxiter );
    cutilCheckMsg("Analysiskernel execution failed");

    unsigned int result = 0;
    cutilSafeCall(cudaMemcpy( &result, d_result,    1 * sizeof(unsigned int), cudaMemcpyDeviceToHost ));
    fprintf(stderr, "working threads: %d out of %d     \r", result, nSeeds);

    Buddhakernel<<<grid, nThreads, shared>>>( accumdata,
                                              coords, lorbit,
                                              rseed, iseed,
                                              iw, iw, ih, TILEH,
                                              clear,
                                              zrmin, zrmax, zimin, zimax,
                                              maxiter, maxR, maxG, maxB,
                                              colormode);
    cutilCheckMsg("Buddhakernel execution failed");

    shared = TILEW*TILEH*sizeof(unsigned int);
    HDR2LDRKernel<<<grid, nThreads, shared>>>( accumdata,
                                               overflowdata,
                                               displaydata,
                                               iw, iw, ih, TILEH,
                                               offset, scale );
    cutilCheckMsg("HDR2LDRKernel execution failed");

    if (denoise)
    {
        cutilSafeCall(cudaMemcpyToArray(array, 0, 0, displaydata, Bpp*sizeof(Pixel)*iw*ih, cudaMemcpyDeviceToDevice));

        cutilSafeCall(cudaBindTextureToArray(tex, array));

        dim3 threads(BLOCKDIM_X, BLOCKDIM_Y);
        dim3 grid(iDivUp(iw, BLOCKDIM_X), iDivUp(ih, BLOCKDIM_Y));

        NLM2<<<grid, threads>>>(displaydata, iw, ih, 0.47f, 0.2f);

        cutilSafeCall(cudaUnbindTexture(tex));
    }
}
