/*
* Copyright (c) 2008-2009 Christian Buchner <Christian.Buchner@gmail.com>
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY Christian Buchner ''AS IS'' AND ANY 
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Christian Buchner BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef CUDACOMPLEX_H
#define CUDACOMPLEX_H

#include <vector_types.h>  // required for float2, double2

#include "cudaconventions.h"  // define my own CUDA coding conventions

#include "doublesingle.h"

/**
 * A templated complex number type for use with CUDA.
 * V2 is the CUDA vector type for storage of type S, where S is the scalar type.
 * This is deliberately designed to use few C++ features in order to work with most
 * CUDA SDK versions. It is friendlier to use than the cuComplex type because it
 * provides more operator overloads.
 * The class should work in host code and in device code and also in emulation mode.
 * Also this has been tested on any OS that the CUDA SDK is available for.
 */
template <typename V2, typename S> struct _cudacomplex {

    // S is the scalar type. With this typedef we make it accessible from any
    // class that uses _cudacomplex<...> as a template parameter.
    typedef S scalar;

    // float2  is a native CUDA type and allows for coalesced 128 byte access
    // double2 is a native CUDA type and allows for coalesced 256 byte access
    // when accessed according to CUDA's memory coalescing rules.
    // x member is real component
    // y member is imaginary component
    V2 value;

#if 0
    _cudacomplex::_cudacomplex() {
    };

    _cudacomplex::_cudacomplex(const double REF(a)) : value( V2(a,0) ) {
    };
#endif

    // assignment of a scalar to complex
    _cudacomplex<V2, S>& operator=(const S REF(a)) {
       value.x = a; value.y = 0;
       return *this;
    };

    // assignment of a pair of Ts to complex
    _cudacomplex<V2, S>& operator=(const S ARRAYREF(a,2)) {
       value.x = a[0]; value.y = a[1];
       return *this;
    };

    // return references to the real and imaginary components
    M_HOSTDEVICE S& real() {return value.x;};
    M_HOSTDEVICE S& imag() {return value.y;};

    // compare two complex numbers
    M_HOSTDEVICE bool operator == (const _cudacomplex<V2, S> REF(b)) const {
        return value.x == b.value.x && value.y == b.value.y;
    }

    // add complex numbers
    M_HOSTDEVICE _cudacomplex<V2, S> operator+(const _cudacomplex<V2, S> REF(b)) const {
        _cudacomplex<V2, S> result = {{ value.x + b.value.x, value.y  + b.value.y }};
        return result;
    }

    // add scalar to complex
    M_HOSTDEVICE _cudacomplex<V2, S> operator+(const S REF(b)) const {
        _cudacomplex<V2, S> result = {{ value.x + b, value.y }};
        return result;
    }

    // subtract complex numbers
    M_HOSTDEVICE _cudacomplex<V2, S> operator-(const _cudacomplex<V2, S> REF(b)) const {
        _cudacomplex<V2, S> result = {{ value.x - b.value.x, value.y  - b.value.y }};
        return result;
    }

    // negate a complex number
    M_HOSTDEVICE _cudacomplex<V2, S> operator-() const {
        _cudacomplex<V2, S> result = {{ -value.x, -value.y }};
        return result;
    }

    // subtract scalar from complex
    M_HOSTDEVICE _cudacomplex<V2, S> operator-(const S REF(b)) const {
        _cudacomplex<V2, S> result = {{ value.x - b, value.y }};
        return result;
    }

    // multiply complex numbers
    M_HOSTDEVICE _cudacomplex<V2, S> operator*(const _cudacomplex<V2, S> REF(b)) const {
        _cudacomplex<V2, S> result = {{ value.x * b.value.x - value.y * b.value.y,
                                        value.y * b.value.x + value.x * b.value.y }};
        return result;
    }

    // multiply complex with scalar
    M_HOSTDEVICE _cudacomplex<V2, S> operator*(const S REF(b)) const {
        _cudacomplex<V2, S> result = {{ value.x * b, value.y * b }};
        return result;
    }

    // divide complex numbers
    M_HOSTDEVICE _cudacomplex<V2, S> operator/(const _cudacomplex<V2, S> REF(b)) const {
        S tmp = ( b.value.x * b.value.x + b.value.y * b.value.y );
        _cudacomplex<V2, S> result = {{ (value.x * b.value.x + value.y * b.value.y ) / tmp,
                                        (value.y * b.value.x - value.x * b.value.y ) / tmp }};
        return result;
    }

    // divide complex by scalar
    M_HOSTDEVICE _cudacomplex<V2, S> operator/(const S REF(b)) const {
        _cudacomplex<V2, S> result = {{ value.x / b, value.y / b }};
        return result;
    }

    // complex conjugate
    M_HOSTDEVICE _cudacomplex<V2, S> operator~() const {
        _cudacomplex<V2, S> result = {{ value.x, -value.y }};
        return result;
    }

    // complex modulus (complex absolute)
    M_HOSTDEVICE S abs() {
        S result = sqrt( value.x*value.x + value.y*value.y );
        return result;
    }

    // complex modulus squared (complex absolute squared)
    M_HOSTDEVICE S abs2() {
        S result = value.x*value.x + value.y*value.y;
        return result;
    }

    // complex phase angle
    M_HOSTDEVICE S phase() {
        S result = __atan2( value.y, value.x );
        return result;
    }

    // a possible alternative to a _cudacomplex constructor
    static M_HOSTDEVICE _cudacomplex<V2, S> make(S a, S b)
    {
        _cudacomplex<V2, S> res;
        res.real() = a;
        res.imag() = b;
        return res;
    }

    // return constant number one
    static M_HOSTDEVICE const _cudacomplex<V2, S> one() {
        return make((S)1.0, (S)0.0);
    }

    // return constant number zero
    static M_HOSTDEVICE const _cudacomplex<V2, S> zero() {
        return make((S)0.0, (S)0.0);
    }

    // return constant number I
    static M_HOSTDEVICE const _cudacomplex<V2, S> I() {
        return make((S)0.0, (S)1.0);
    }
    
};


//
// Define the common complex types
//
typedef ALIGN(8) _cudacomplex<float2, float> singlecomplex;
typedef ALIGN(16) _cudacomplex<double2, double> doublecomplex;
typedef ALIGN(16) _cudacomplex<doublesingle2, doublesingle> doublesinglecomplex;


//
// Non-member overloads for single complex
//

// subtract single complex from scalar
HOSTDEVICE _cudacomplex<float2, float> operator-(const float REF(a), const _cudacomplex<float2, float> REF(b)) {
    _cudacomplex<float2, float> result = {{ a - b.value.x, -b.value.y }};
    return result;
}

// add single complex to scalar
HOSTDEVICE _cudacomplex<float2, float> operator+(const float REF(a), const _cudacomplex<float2, float> REF(b)) {
    _cudacomplex<float2, float> result = {{ a + b.value.x, b.value.y }};
    return result;
}

// multiply scalar with single complex
HOSTDEVICE _cudacomplex<float2, float> operator*(const float REF(a), const _cudacomplex<float2, float> REF(b)) {
    _cudacomplex<float2, float> result = {{ a * b.value.x, a * b.value.y }};
    return result;
}

// divide scalar by single complex
HOSTDEVICE _cudacomplex<float2, float> operator/(const float REF(a), const _cudacomplex<float2, float> REF(b)) {
    float tmp = ( b.value.x * b.value.x + b.value.y * b.value.y );
    _cudacomplex<float2, float> result = {{ ( a * b.value.x ) / tmp, ( -a * b.value.y ) / tmp }};
    return result;
}


//
// Non-member overloads for double complex
//

// subtract double complex from scalar
HOSTDEVICE _cudacomplex<double2, double> operator-(const double REF(a), const _cudacomplex<double2, double> REF(b)) {
    _cudacomplex<double2, double> result = {{ a - b.value.x, -b.value.y }};
    return result;
}

// add double complex to scalar
HOSTDEVICE _cudacomplex<double2, double> operator+(const double REF(a), const _cudacomplex<double2, double> REF(b)) {
    _cudacomplex<double2, double> result = {{ a + b.value.x, b.value.y }};
    return result;
}

// multiply scalar with double complex
HOSTDEVICE _cudacomplex<double2, double> operator*(const double REF(a), const _cudacomplex<double2, double> REF(b)) {
    _cudacomplex<double2, double> result = {{ a * b.value.x, a * b.value.y }};
    return result;
}

// divide scalar by double complex
HOSTDEVICE _cudacomplex<double2, double> operator/(const double REF(a), const _cudacomplex<double2, double> REF(b)) {
    double tmp = ( b.value.x * b.value.x + b.value.y * b.value.y );
    _cudacomplex<double2, double> result = {{ ( a * b.value.x ) / tmp, ( -a * b.value.y ) / tmp }};
    return result;
}


//
// Non-member overloads for doublesingle complex
//

// subtract doublesingle complex from scalar
HOSTDEVICE _cudacomplex<doublesingle2, doublesingle> operator-(const doublesingle REF(a), const _cudacomplex<doublesingle2, doublesingle> REF(b)) {
    _cudacomplex<doublesingle2, doublesingle> result = {{ a - b.value.x, -b.value.y }};
    return result;
}

// add doublesingle complex to scalar
HOSTDEVICE _cudacomplex<doublesingle2, doublesingle> operator+(const doublesingle REF(a), const _cudacomplex<doublesingle2, doublesingle> REF(b)) {
    _cudacomplex<doublesingle2, doublesingle> result = {{ a + b.value.x, b.value.y }};
    return result;
}

// multiply scalar with doublesingle complex
HOSTDEVICE _cudacomplex<doublesingle2, doublesingle> operator*(const doublesingle REF(a), const _cudacomplex<doublesingle2, doublesingle> REF(b)) {
    _cudacomplex<doublesingle2, doublesingle> result = {{ a * b.value.x, a * b.value.y }};
    return result;
}

// divide scalar by doublesingle complex
HOSTDEVICE _cudacomplex<doublesingle2, doublesingle> operator/(const doublesingle REF(a), const _cudacomplex<doublesingle2, doublesingle> REF(b)) {
    doublesingle tmp = ( b.value.x * b.value.x + b.value.y * b.value.y );
    _cudacomplex<doublesingle2, doublesingle> result = {{ ( a * b.value.x ) / tmp, ( -a * b.value.y ) / tmp }};
    return result;
}


// a possible alternative to a single complex constructor
HOSTDEVICE singlecomplex make_singlecomplex(float a, float b)
{
    singlecomplex res;
    res.real() = a;
    res.imag() = b;
    return res;
}


// a possible alternative to a double complex constructor
HOSTDEVICE doublecomplex make_doublecomplex(double a, double b)
{
    doublecomplex res;
    res.real() = a;
    res.imag() = b;
    return res;
}

// a possible alternative to a doublesingle complex constructor
HOSTDEVICE doublesinglecomplex make_doublesinglecomplex(doublesingle a, doublesingle b)
{
    doublesinglecomplex res;
    res.real() = a;
    res.imag() = b;
    return res;
}

#endif // #ifndef CUDACOMPLEX_H
